from visidata import BaseSheet

# for mac users to use Option+x as Alt+x without reconfiguring the terminal

# Option+X
BaseSheet.bindkey('å', 'Alt+a')
BaseSheet.bindkey('∫', 'Alt+b')
BaseSheet.bindkey('ç', 'Alt+c')
BaseSheet.bindkey('∂', 'Alt+d')
BaseSheet.bindkey('´', 'Alt+e')
BaseSheet.bindkey('ƒ', 'Alt+f')
BaseSheet.bindkey('©', 'Alt+g')
BaseSheet.bindkey('˙', 'Alt+h')
BaseSheet.bindkey('ˆ', 'Alt+i')
BaseSheet.bindkey('∆', 'Alt+j')
BaseSheet.bindkey('˚', 'Alt+k')
BaseSheet.bindkey('¬', 'Alt+l')
BaseSheet.bindkey('µ', 'Alt+m')
BaseSheet.bindkey('˜', 'Alt+n')
BaseSheet.bindkey('ø', 'Alt+o')
BaseSheet.bindkey('π', 'Alt+p')
BaseSheet.bindkey('œ', 'Alt+q')
BaseSheet.bindkey('®', 'Alt+r')
BaseSheet.bindkey('ß', 'Alt+s')
BaseSheet.bindkey('†', 'Alt+t')
BaseSheet.bindkey('¨', 'Alt+u')
BaseSheet.bindkey('√', 'Alt+v')
BaseSheet.bindkey('∑', 'Alt+w')
BaseSheet.bindkey('≈', 'Alt+x')
BaseSheet.bindkey('¥', 'Alt+y')
BaseSheet.bindkey('Ω', 'Alt+z')

# Option+Shift+X
BaseSheet.bindkey('Å', 'Alt+A')
BaseSheet.bindkey('ı', 'Alt+B')
BaseSheet.bindkey('Ç', 'Alt+C')
BaseSheet.bindkey('Î', 'Alt+D')
#BaseSheet.bindkey('´', 'Alt+E')
BaseSheet.bindkey('Ï', 'Alt+F')
BaseSheet.bindkey('˝', 'Alt+G')
BaseSheet.bindkey('Ó', 'Alt+H')
#BaseSheet.bindkey('ˆ', 'Alt+I')
BaseSheet.bindkey('Ô', 'Alt+J')
BaseSheet.bindkey('', 'Alt+K')  # apple logo
BaseSheet.bindkey('Ò', 'Alt+L')
BaseSheet.bindkey('Â', 'Alt+M')
#BaseSheet.bindkey('˜', 'Alt+N')
BaseSheet.bindkey('Ø', 'Alt+O')
BaseSheet.bindkey('∏', 'Alt+P')
BaseSheet.bindkey('Œ', 'Alt+Q')
BaseSheet.bindkey('‰', 'Alt+R')
BaseSheet.bindkey('Í', 'Alt+S')
BaseSheet.bindkey('ˇ', 'Alt+T')
#BaseSheet.bindkey('¨', 'Alt+U')
BaseSheet.bindkey('◊', 'Alt+V')
BaseSheet.bindkey('„', 'Alt+W')
BaseSheet.bindkey('˛', 'Alt+X')
BaseSheet.bindkey('Á', 'Alt+Y')
BaseSheet.bindkey('¸', 'Alt+Z')

# Option+nonletter
BaseSheet.bindkey('¡', 'Alt+1')
BaseSheet.bindkey('™', 'Alt+2')
BaseSheet.bindkey('£', 'Alt+3')
BaseSheet.bindkey('¢', 'Alt+4')
BaseSheet.bindkey('∞', 'Alt+5')
BaseSheet.bindkey('§', 'Alt+6')
BaseSheet.bindkey('¶', 'Alt+7')
BaseSheet.bindkey('•', 'Alt+8')
BaseSheet.bindkey('ª', 'Alt+9')
BaseSheet.bindkey('º', 'Alt+0')
BaseSheet.bindkey('', 'Alt+`')
BaseSheet.bindkey('–', 'Alt+-')
BaseSheet.bindkey('≠', 'Alt+=')
BaseSheet.bindkey('“', 'Alt+[')
BaseSheet.bindkey('‘', 'Alt+]')
BaseSheet.bindkey('«', 'Alt+\\')
BaseSheet.bindkey('÷', 'Alt+/')
BaseSheet.bindkey('…', 'Alt+;')
BaseSheet.bindkey('æ', 'Alt+\'')
BaseSheet.bindkey('≤', 'Alt+,')
BaseSheet.bindkey('≥', 'Alt+.')

# Option+Shift+nonletter
BaseSheet.bindkey('⁄', 'Alt+!')
BaseSheet.bindkey('€', 'Alt+@')
BaseSheet.bindkey('‹', 'Alt+#')
BaseSheet.bindkey('›', 'Alt+$')
BaseSheet.bindkey('ﬁ', 'Alt+%')
BaseSheet.bindkey('ﬂ', 'Alt+^')
BaseSheet.bindkey('‡', 'Alt+&')
BaseSheet.bindkey('°', 'Alt+*')
BaseSheet.bindkey('·', 'Alt+(')
BaseSheet.bindkey('‚', 'Alt+)')
#BaseSheet.bindkey('`', 'Alt+~')
BaseSheet.bindkey('—', 'Alt+_')
BaseSheet.bindkey('±', 'Alt++')
BaseSheet.bindkey('”', 'Alt+{')
BaseSheet.bindkey('’', 'Alt+}')
BaseSheet.bindkey('»', 'Alt+|')
BaseSheet.bindkey('¿', 'Alt+?')
BaseSheet.bindkey('Ú', 'Alt+:')
BaseSheet.bindkey('Æ', 'Alt+"')
BaseSheet.bindkey('¯', 'Alt+<')
BaseSheet.bindkey('˘', 'Alt+>')
